// -----------------------------
// ------- C++ INCLUDES --------
#define _USE_MATH_DEFINES
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <math.h>
#include <string>

// ------------------------------
// -------- CUDA INCLUDES -------
#include <cuda_runtime.h>
#include <device_launch_parameters.h>
#include <math_constants.h>
#include <cufft.h>
#include <cublas.h>
#include <cublas_v2.h>

// -------------------------------
// ------- MATLAB INCLUDES -------
#include <matrix.h>
#include <mex.h>

using namespace std;

// --------------------------------
// --- --- Global Variables --- ---
static struct Variables{
	int N			= 0; // Time Resolution
	int NumRows		= 0; // Frequency Resolution
	int SwiftStep	= 0; // Kernel Iteration Step
	int FStart		= 0; // Frequency Search Start [Hz]
	int FStop		= 0; // Frequency Search Stop [Hz]
	int FDiff		= 0; // Frequency Step [Hz]
	int Fsa			= 0; // Sampling Frequency
	int Stage		= 0; // Computing Stage
	int FDInterv	= 0; // Total Frequency Space

	dim3 THREADS_PER_BLOCK	= dim3(1, 1, 1);
	dim3 BLOCKS_PER_GRID	= dim3(1, 1, 1);

}Vars;

// -------------------------------
// --- --- DEVICE POINTERS --- ---
static struct DevicePtx{
	cublasHandle_t cublas = NULL;	// Cublas Handle
	cudaEvent_t start = NULL;		// Time Measure Start
	cudaEvent_t stop = NULL;		// Time Measure Stop

	cufftComplex *GPS = NULL;		// Received GPS Signal
	cufftReal* Prn = NULL;			// Generated PRN Replica

	cufftComplex *Out = NULL;
	cufftReal *Help = NULL;

}Dev;


// ----------------------------------------
// -------- Device Complex Multiply -------
__device__ cufftComplex C_Mul(cufftComplex a, cufftComplex b){
	cufftComplex temp;
	temp.x = a.x*b.x - a.y*b.y;
	temp.y = a.y*b.x + b.y*a.x;
	return temp;
}

// ------------------------------------
// -------- Device Complex Add --------
__device__ cufftComplex C_Add(cufftComplex a, cufftComplex b){
	cufftComplex c;
	c.x = a.x + b.x;
	c.y = a.y + b.y;
	return c;
}

// ---------------------------------------------
// ------- Device Complex absolute value -------
__device__ cufftReal C_Abs(cufftComplex A){
	cufftReal res = sqrtf(A.x*A.x + A.y*A.y);
	return res;
}


/*****************************************************
 * @brief: GPU Integration Kernel (Maximum Likelihood)  
 ****************************************************/
__global__ void GPS_ML(cufftComplex *Output, cufftComplex *GPS, float*Prn, int step,int N,float Tsa,int FDiff,int SwiftStep, int Fstart){
	unsigned int tix = threadIdx.x + blockIdx.x*blockDim.x;   // X - Dimension Index
	unsigned int tiy = threadIdx.y + blockIdx.y*blockDim.y;   // Y - Dimension Index

	// --------------------------------------
	// ------- Check Processing Ranges ------
	if (tix < N){
		int index = N*tiy + tix;

		// ----------------------------------------------------
		// ------- Load Intermediate Results from memory ------
		cufftComplex res = Output[index];
		
		// What are we calculating?
		int istart = step*SwiftStep;
		int istop = istart + SwiftStep;
		if (istop > N){ istop = N; };

		// ---------------------------------------------------------
		// ------- Main PieceWise (Chunk) computing routine --------
		cufftReal ConstX = -2.0f *  CUDART_PI_F * Tsa * ((((float)FDiff) * ((float)tiy))  + ((float)Fstart));
		for (int i = istart; i < istop; i++){
			cufftComplex temp;
			cufftReal t1, pr;

			int PrnInd = i - tix;if (PrnInd < 0){ PrnInd = N + PrnInd; } //Fix negative indexing
			pr = Prn[PrnInd];
			t1 = ConstX*i;
			temp.x = cosf(t1) * pr;
			temp.y = sinf(t1) * pr;
			res = C_Add(res, C_Mul(GPS[i], temp));
		}

		// ---------------------------------------------------------
		// ------- Save result/partial result back to Global -------
		Output[index] = res;

	}
}

/********************************************************
 * @brief Computes Absolute Value out of a Complex Number
 *******************************************************/
__global__ void Make_ABS(cufftComplex *in, float *out,int Nx,int Ny){
	unsigned int tix = threadIdx.x + blockIdx.x*blockDim.x;   // X - Dimension Index
	unsigned int tiy = threadIdx.y + blockIdx.y*blockDim.y;   // Y - Dimension Index

	// --------------------------------------
	// ------- Check Processing Ranges ------
	if ((tix < Nx) && (tiy < Ny)){
		int index = Nx*tiy + tix;

		// Compute and Save result to Global Memory
		out[index] = C_Abs(in[index]);
	}

}

//Matlab compatibility
__global__ void Row2ColKernel(cufftReal *in, cufftReal *out, int Nx, int Ny)
{
	int tix = threadIdx.x + blockIdx.x*blockDim.x;
	int tiy = threadIdx.y + blockIdx.y*blockDim.y;
	if ((tix < Nx) && (tiy < Ny)){
		int tin = tiy*Nx + tix;
		int tou = tix*Ny + tiy;
		out[tou] = in[tin];
	}
}


/***************************
* @brief CUDA ERROR HANDLER
***************************/
#define HANDLE_ERROR( err ) (HandleError( err, __FILE__, __LINE__ ))
static void HandleError(cudaError_t err, const char* file, int line) {
	if (err != cudaSuccess) {
		std::string str = std::string("CUDA ERROR: %s in %s at line %d\n") + cudaGetErrorString(err) + std::string(file) + std::to_string(line);
		mexErrMsgTxt(str.c_str());
	}
}

/********************************************
 *@brief: Allocates Resources for Computation 
 ********************************************/
void AllocateResources(){

	if (Dev.start		== NULL){ HANDLE_ERROR(cudaEventCreate(&Dev.start)); }
	if (Dev.stop		== NULL){ HANDLE_ERROR(cudaEventCreate(&Dev.stop)); }
	if (Dev.GPS			== NULL){ HANDLE_ERROR(cudaMalloc((void**)&Dev.GPS, Vars.N*sizeof(cufftComplex))); }
	if (Dev.Out			== NULL){ HANDLE_ERROR(cudaMalloc((void**)&Dev.Out, Vars.N*Vars.NumRows*sizeof(cufftComplex))); }
	if (Dev.Prn			== NULL){ HANDLE_ERROR(cudaMalloc((void**)&Dev.Prn, Vars.N*sizeof(cufftReal))); }
	if (Dev.Help		== NULL){ HANDLE_ERROR(cudaMalloc((void**)&Dev.Help, Vars.N*Vars.NumRows*sizeof(cufftReal))); }
	if (Dev.cublas		== NULL){ cublasCreate(&Dev.cublas); }
}

/***********************************
 *@brief: Cleans Allocated Resources 
 ***********************************/
void CleanResources(){

	if (Dev.GPS			!= NULL){ HANDLE_ERROR(cudaFree(Dev.GPS)); Dev.GPS = NULL; }
	if (Dev.Out			!= NULL){ HANDLE_ERROR(cudaFree(Dev.Out)); Dev.Out = NULL; }
	if (Dev.Prn			!= NULL){ HANDLE_ERROR(cudaFree(Dev.Prn)); Dev.Prn = NULL; }
	if (Dev.Help		!= NULL){ HANDLE_ERROR(cudaFree(Dev.Help)); Dev.Help = NULL; }
	if (Dev.stop		!= NULL){ HANDLE_ERROR(cudaEventDestroy(Dev.stop)); Dev.stop = NULL; }
	if (Dev.start		!= NULL){ HANDLE_ERROR(cudaEventDestroy(Dev.start)); Dev.start = NULL; }
	if (Dev.cublas		!= NULL){ cublasDestroy(Dev.cublas); Dev.cublas = NULL; }

}

/*************************************************
 * @brief: Calculates the ML of a GPS Signal.
 * - Estimates the Signal Delay and Doppler Offset
 *************************************************/
void Calculate(mxSingle*Est_Doppler, mxSingle*Est_Delay, mxSingle*Est_Time,mxSingle *Space_FreqTime,mxComplexSingle *GPS_Sig,mxSingle *GPS_Prn){
	cufftReal ElapsedTime = -1.0f;
	HANDLE_ERROR(cudaEventRecord(Dev.start));



	// -------------------------------------------------------------------
	// ------- Copy Received GPS I/Q Signal and PRN Replica to GPU ------- 
	HANDLE_ERROR(cudaMemcpy(Dev.Prn, GPS_Prn, Vars.N*sizeof(cufftReal), cudaMemcpyHostToDevice));
	HANDLE_ERROR(cudaMemcpy(Dev.GPS, GPS_Sig, Vars.N*sizeof(cufftComplex), cudaMemcpyHostToDevice));

	// --------------------------------------
	// ------ Reset to Zero By default ------
	cudaMemset(Dev.Out, 0x0, Vars.N * Vars.NumRows * sizeof(cufftComplex));


	// -----------------------------------
	// ------- Main Computing Loop -------
	float Tsa = 1.0f / static_cast<float>(Vars.Fsa); // Sampling Period
	for (int i = 0; i < (Vars.N / Vars.SwiftStep + 1); i++){
		GPS_ML << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> >(Dev.Out, Dev.GPS, Dev.Prn, i, Vars.N, Tsa, Vars.FDiff, Vars.SwiftStep, Vars.FStart);
		HANDLE_ERROR(cudaGetLastError());
	}

	// ---------------------------------------
	// ------- Compute Absolute Values -------
	Make_ABS << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> >(Dev.Out, Dev.Help, Vars.N, Vars.NumRows);
	HANDLE_ERROR(cudaGetLastError());

	// ---------------------------------------------
	// ------- CONVERT TO COLUMN MAJOR ORDER -------
	Row2ColKernel << <Vars.BLOCKS_PER_GRID, Vars.THREADS_PER_BLOCK >> >(Dev.Help, (cufftReal*)Dev.Out, Vars.N, Vars.NumRows);
	HANDLE_ERROR(cudaGetLastError());

	// ----------------------------------------------------------
	// ------- Copy Time-Freq Space back to Host (Matlab) -------
	HANDLE_ERROR(cudaMemcpy(Space_FreqTime, Dev.Out, Vars.N * Vars.NumRows * sizeof(cufftReal), cudaMemcpyDeviceToHost));

	//SEARCH FOR PEAK USING CUBLAS
	int Index = -1;
	cublasIsamax(Dev.cublas, Vars.N*Vars.NumRows, (cufftReal*)Dev.Out, 1, &Index);

	//GET Delay and Fdop offset
	int Tind = static_cast<int>(std::fmod(static_cast<double>(Index), static_cast<double>(Vars.N)));
	int Find = (Index - Tind) / Vars.N;
	cufftReal Doppler = static_cast<cufftReal>(Find*Vars.FDiff - Vars.FStart);
	cufftReal Delay = static_cast<cufftReal>(Tind);

	// -----------------------------------
	// ------ GET CALCULATION TIME -------
	HANDLE_ERROR(cudaEventRecord(Dev.stop));
	HANDLE_ERROR(cudaEventSynchronize(Dev.stop));
	HANDLE_ERROR(cudaEventElapsedTime(&ElapsedTime, Dev.start, Dev.stop));

	// ---------------------------------------------------
	// ------- Assign Estimations to Matlab Output -------
	Est_Time[0]		= static_cast<float>(ElapsedTime);
	Est_Delay[0]	= static_cast<float>(Delay);
	Est_Doppler[0]	= static_cast<float>(Doppler);

	return;
}


/******************************************************
 *@brief: Matlab Entry Function
 * prhs = Pointer to The Right Hand Side	==> INPUT
 * plhs = Pointer to The Left Hand Side		==> OUTPUT
 ******************************************************/
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]){


	// -------------------------------------
	// ------ CleanUp Used Resources -------
	if (nrhs == 1){
		CleanResources();
		mexPrintf("CUDA Resources Cleaned\n");
		return;
	}

	// --------------------------
	// ----- Input Arrays -------
	const mxArray* GPS		= prhs[0]; // Received Signal
	const mxArray* PRN		= prhs[1]; // Generated PRN Sequence
	const mxArray* AddIn	= prhs[2]; // Computing Parameters

	// --------------------------------------
	// ------- Check for Correct Input ------
	if (!mxIsComplex(GPS)) {mexErrMsgTxt("Error GPS Signal needs to be Complex!\n");}
	if (mxIsComplex(PRN))  {mexErrMsgTxt("Error PRN Signal needs to be Real!\n");}
	if (mxGetClassID(GPS) != mxSINGLE_CLASS) { mexErrMsgTxt("Error GPS sequence must be in single precision!\n"); }
	if (mxGetClassID(PRN) != mxSINGLE_CLASS) { mexErrMsgTxt("Error PRN sequence must be in single precision!\n"); }

	// -----------------------------
	// ------- Grab Pointers -------
	mxComplexSingle *GPS_Sig = mxGetComplexSingles(GPS);
	mxSingle		*GPS_Prn = mxGetSingles(PRN);

	// ------------------------------------------
	// ----- Extract Computing Parameters ------- 
	double* Add = mxGetPr(AddIn);
	int Old_FStart		= Vars.FStart;
	int Old_FStop		= Vars.FStop;
	int Old_FDiff		= Vars.FDiff;
	int Old_NumSamples	= Vars.N;
	Vars.FStart		= static_cast<int>(Add[0]);		// Doppler frequency start
	Vars.FStop		= static_cast<int>(Add[1]);     // Doppler frequency stop
	Vars.FDiff		= static_cast<int>(Add[2]);     // Frequency scan step
	Vars.N			= static_cast<int>(Add[3]);     // Length of the GPS signal
	Vars.Fsa		= static_cast<int>(Add[4]);     // GPS Sampling Frequncy
	Vars.SwiftStep	= static_cast<int>(Add[5]);		// Batch Calculation
	Vars.Stage		= static_cast<int>(Add[6]);     // Init/Kill/Compute

	// -------------------------------------------------
	// ------- Reallocate Resources if Necessary -------
	if ((Old_FStart != Vars.FStart) || (Old_FStop != Vars.FStop) || (Old_NumSamples != Vars.N) || (Old_FDiff != Vars.FDiff)){

		// ------------------------------------------------------
		// ------- Cleanup Previously Allocated Resources -------
		CleanResources();

		// ----------------------------------------------------------
		// Computing Grid Update :
		// We will be looking for a maximum in time-frequency space.
		// --> Time Delay is bound by amount of Samples
		// --> Frequency is bound by Start / Stop Ranges and Freq Step.
		//		For 1M Samples and Frequency in range -6000:2:7000,there will be a total of 1M * (6000 + 7000)/2 + 1 points.
		Vars.FDInterv = std::abs(Vars.FStart) + std::abs(Vars.FStop);
		Vars.NumRows = (Vars.FDInterv) / Vars.FDiff + 1;


		// -------------------------------------------------
		// ------ Specify Kernel Launching Dimensions ------
		Vars.THREADS_PER_BLOCK = dim3(768, 1, 1);
		Vars.BLOCKS_PER_GRID = dim3(Vars.N / 768 + 1, Vars.FDInterv / Vars.FDiff + 1, 1);

		// -----------------------------------------------------
		// -------- Allocate Parameters with New Settings ------
		AllocateResources();
	}

	// ----------------------------------------
	// ------- Associate Matlab Output --------
	size_t outDims[2] = { static_cast<size_t>(Vars.NumRows), static_cast<size_t>(Vars.N) };
	size_t simple[2] = { 1, 1 };
	mxArray *MatlabOut	= plhs[0] = mxCreateNumericArray(2, outDims,	mxSINGLE_CLASS, mxREAL);	// 
	mxArray *Doppler	= plhs[1] = mxCreateNumericArray(2, simple,		mxSINGLE_CLASS, mxREAL);	// Doppler Offset (1x1 Real)
	mxArray *Delay		= plhs[2] = mxCreateNumericArray(2, simple,		mxSINGLE_CLASS, mxREAL);	// Delay   Offset (1x1 Real)
	mxArray *ElTime		= plhs[3] = mxCreateNumericArray(2, simple,		mxSINGLE_CLASS, mxREAL);	// Elapset Time   (1x1 Real)
	
	mxSingle* M_Doppler		= mxGetSingles(Doppler);
	mxSingle* M_Delay		= mxGetSingles(Delay);
	mxSingle* M_Time		= mxGetSingles(ElTime);
	mxSingle* M_FreqTime	= mxGetSingles(MatlabOut);


	//SWITCH COMPUTATION TYPE
	switch (Vars.Stage){
		case 0:{ 
			AllocateResources();
			break;
		}
		case 1:{ //COMPUTE
			Calculate(M_Doppler, M_Delay, M_Time, M_FreqTime, GPS_Sig, GPS_Prn);
			break;
		}
		default:{
			break;
		}
	}


	return;
}



