function Callback_GNSS_Compute(hObject,callbackdata)
    global S;
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%% Parse Data from UI Boxes %%%%%%%
    Delay       = str2double(get(S.EditDelay,'String'));
    Phase       = str2double(get(S.EditPhase,'String'));
    Doppler     = str2double(get(S.EditOffset,'String'));
    Sweep       = str2double(get(S.EditSweep,'String'));
    FStart      = str2double(get(S.EditFStart,'String'));
    FStop       = str2double(get(S.EditFStop,'String'));
    SNR         = str2double(get(S.EditSNR,'String'));
    GPSGen      = str2double(get(S.EditGPSCod,'String'));
    GPSDec      = str2double(get(S.EditGPSDec,'String'));
    Fsa         = str2double(get(S.EditFsa,'String'));
    Time        = str2double(get(S.EditTime,'String'));
    FreqPoints  = str2double(get(S.EditFPoints,'String'));

    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%% Check for Invalid User Input %%%%%%
    Values = [Delay,Phase,Doppler,Sweep,FStart,FStop,SNR,GPSGen,GPSDec,Fsa,Time,FreqPoints];
    if(any(isnan(Values)) || any(isinf(Values)))
      msgbox('Invalid Input, please try again','Parsing Error','error');
        return;
    end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%% Convert Data to Required Precision %%%%%%%
    Delay       = round(Delay);
    Sweep       = round(Sweep);
    FStart      = round(FStart); 
    FStop       = round(FStop);
    GPSGen      = round(GPSGen);
    GPSDec      = round(GPSDec);
    Fsa         = round(Fsa);
    FreqPoints  = round(FreqPoints);
    Phase       = deg2rad(Phase);
    Doppler     = Doppler;
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%% Validate Input %%%%%%%
    if(GPSGen < 1 || GPSGen > 37)
        msgbox('GPS Generation PRN Should be from interval: <1,37>',S.MsgTitle,'warn');
        return;
    end

    if(GPSDec < 1 || GPSDec > 37)
        msgbox('GPS Decode PRN Should be from interval: <1,37>',S.MsgTitle,'warn');
        return;
    end

    if(Delay < 0)
        msgbox('Delay cannot be negative',S.MsgTitle,'warn');
        return;
    end
    if(FStart > FStop)
        msgbox('Frequency Sweep Start must be lower than Frequency Sweep Stop',S.MsgTitle,'warn');
        return;
    end

    if(Sweep < 1)
        msgbox('Sweeping Frequency must be at least 1 Hz',S.MsgTitle,'warn');
        return;
    end
    
    if(FreqPoints < 1)
        msgbox('Number of samples in Freq. domain must be > 1 !',S.MsgTitle,'warn');
        return;
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%% Generate TIme Axis %%%%%%%
    t = 0:1/Fsa:Time - 1/Fsa;
    N = length(t);
    
    if(Delay > N)
        msgbox(['Delay cannot be greater than ',num2str(N)],S.MsgTitle,'warn');
        return;
    end
    
    if(N < 1023)
        msgbox('Need More samples, Increase either integration time, or sampling frequency',S.MsgTitle,'warn');
        return;
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%% Update Boxes with Used Values %%%%%%%
    set(S.EditDelay,'String',num2str(Delay));
    set(S.EditPhase,'String',num2str(rad2deg(Phase)));
    set(S.EditOffset,'String',num2str(Doppler));
    set(S.EditSweep,'String',num2str(Sweep));
    set(S.EditFStart,'String',num2str(FStart));
    set(S.EditFStop,'String',num2str(FStop));
    set(S.EditSNR,'String',num2str(SNR));
    set(S.EditGPSCod,'String',num2str(GPSGen));
    set(S.EditGPSDec,'String',num2str(GPSDec));
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%% GPS SIGNAL GENERATION %%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    % LOAD USED PRN
    PRN = load(strcat('..',filesep,'Data',filesep,'PRN',num2str(GPSGen),'.txt'));
    PRN = PRN*2 - 1; % 0,1 to -1,1
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%% Resample PRN to Match Time Axis %%%%%%%
    counter = 1;
    coef = N / 1023;
    C = zeros(1,N);
    for i = 1:N
          temp = round(coef*counter);
          if(temp < i)
              counter  = counter +1;
          end
          C(i) = PRN(counter);
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%% Generate GPS Signal %%%%%%%%
    xt = zeros(1,N);
    for i = 1:N;
        PrnIndex = i-Delay;
        if(PrnIndex < 1)
            PrnIndex = N+PrnIndex;
        end
        xt(i) = S.A.*C(PrnIndex)*exp(1i*2*pi*Doppler*t(i) + 1i*Phase);
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%% Add Gaussian Noise %%%%%%%%
    b1 = randn(1,N);
    Pb1 = mean(b1.*b1);             % Noise strength
    Ps1 = mean(abs(xt).*abs(xt));   % Signal strength
    k = sqrt(Ps1/Pb1*10^(-SNR/10)); % Noise scaling coefficient
    Noise = k*b1 + k*1i*b1;
    xt = xt + Noise;
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%% Load Decoding PRN Sequence %%%%%%%%
    if(GPSDec == GPSGen)
        Cdec = C;
    else
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Load and Interpolate the PRN to match time Axis %
        PRNDec = load(strcat('Data',filesep,'PRN',num2str(GPSDec),'.txt'));
        PRNDec = PRNDec*2 - 1; % 0,1 to -1,1
        counter = 1;
        coef = N / 1023;
        Cdec = zeros(1,N);
        for i = 1:N
              temp = round(coef*counter);
              if(temp < i)
                  counter  = counter +1;
              end
              Cdec(i) = PRNDec(counter);
        end
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%% Compute the Time-Frequency 3D Space %%%%%%%
    switch S.ComputingType
        case 'CUDA ML'
            
            % CHECK FOR POSSIBLE CUDA PROCESSSING PROBLEMS
            if(gpuDeviceCount == 0)
                msgbox('NO CUDA Device available!','Error','error');
                return;
            end
            dev = gpuDevice;
            CMajor = fix(str2num(dev.ComputeCapability));
            if((dev.DriverVersion < 6.5) || (CMajor < 2))
                if(dev.DriverVersion < 6.5)
                      msgbox('Cannot proceed, please update nVidia Graphics Drivers!','cudaErrorInsufficientDriver','error');
                    return;
                else
                    msgbox('Requires a CC2+ Graphics processor! (GeForce 400+)','Low Compute Capability','error');
                    return;
                end
            end
            if(~strcmp(computer,'PCWIN64'))
                msgbox('Supports only 64-Bit Matlab','System error','error');
                return;
            end
            % CUDA PROCEED WITH COMPUTATUON
             GPSS = GPSDopplerSim(single(xt),single(Cdec),[FStart,FStop,Sweep,N,Fsa,S.SwiftStep,1]);
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        case 'CPU ML' %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            GPSS = Compute_GNSS_CPU_ML(xt,Cdec,FStart,FStop,Sweep,Fsa);
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        case 'CPU FFT' %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            GPSS = Compute_GNSS_CPU_FFT(xt,Cdec,FStart,FStop,Sweep,t);
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        case 'DBZP' %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            [GPSS,Nc] = Compute_GNSS_CPU_DBZP(xt,Cdec,FreqPoints);
        otherwise
            error('Unknown Computing Type');
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%% Plot Results %%%%%%%%
    switch S.ComputingType
        case 'DBZP'
            surf(S.Ax3D,GPSS,'edgecolor','none');
            set(S.Ax3D.XLabel,'String','Chip Delay [N*Ts]');
            set(S.Ax3D.YLabel,'String','Doppler offset [Hz]');
            
            rot = get(S.Ax3D,'YTickLabelRotation');
            yTicks = linspace(1,FreqPoints,S.NumLabels);
            yTicksReal = linspace((-Fsa/(Nc*2)),(Fsa/(Nc*2)),S.NumLabels);
            for i = 1:S.NumLabels
                yTickLabels(i) = cellstr(num2str(yTicksReal(i)));
            end
            
            set(S.Ax3D,'YTickMode','manual');
            set(S.Ax3D,'YTick',yTicks);
            set(S.Ax3D,'YTickLabel',yTickLabels);
            set(S.Ax3D,'YTickLabelRotation',rot);
            
            [NN,II] = max(GPSS);
            [N,I] = max(NN);
            plot(S.AxDel,0:length(GPSS(1,:))-1,GPSS(II(I),:)','LineWidth',S.Lw);
            set(S.AxDel.XLabel,'String','Delay [N*TS s]');
            set(S.AxDel,'XGrid','on');
            set(S.AxDel,'YGrid','on');
            
            plot(S.AxDop,linspace(-Fsa/(Nc*2),Fsa/(Nc*2),length(GPSS(:,I))),GPSS(:,I),'LineWidth',S.Lw);
            set(S.AxDop.XLabel,'String','Doppler offset [Hz]');
            set(S.AxDop,'XGrid','on');
            set(S.AxDop,'YGrid','on');
            
        otherwise
            surf(S.Ax3D,GPSS,'edgecolor','none');
            set(S.Ax3D.XLabel,'String','Chip Delay [N*Ts]');
            set(S.Ax3D.YLabel,'String','Doppler offset [Hz]');

            rot = get(S.Ax3D,'YTickLabelRotation');
            NumIndexes = length(FStart:Sweep:FStop);
            yTicks = linspace(1,NumIndexes,S.NumLabels);
            yTicksReal = linspace(FStart,FStop,S.NumLabels);
            for i = 1:S.NumLabels
                yTickLabels(i) = cellstr(num2str(yTicksReal(i)));
            end
            set(S.Ax3D,'YTickMode','manual');
            set(S.Ax3D,'YTick',yTicks);
            set(S.Ax3D,'YTickLabel',yTickLabels);
            set(S.Ax3D,'YTickLabelRotation',rot);


            [NN,II] = max(GPSS);
            [N,I] = max(NN);
            plot(S.AxDel,0:length(GPSS(1,:))-1,GPSS(II(I),:),'LineWidth',S.Lw);
            set(S.AxDel.XLabel,'String','Delay [N*TS s]');
            set(S.AxDel,'XGrid','on');
            set(S.AxDel,'YGrid','on');
            
            
            XDopAxis = FStart:Sweep:FStop;
            plot(S.AxDop,XDopAxis,GPSS(:,I),'LineWidth',S.Lw);
            set(S.AxDop.XLabel,'String','Doppler offset [Hz]');
            set(S.AxDop,'XGrid','on');
            set(S.AxDop,'YGrid','on');
    end
    
    
end