%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% @brief: Calculates the Frequency-Delay Space
% xt ...         Input GPS Signal
% C ...          PRN Sequence to decode with
% FreqPoints ... Number of Samples per 1 spectrum (Doppler resolution)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [ GPSSpace, Nc ] =  Compute_GNSS_CPU_DBZP(xt,C,FreqPoints);


%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Check for Invalid Input %
if(length(xt) ~= length(C))
    error('Please, check input signal length and try again');
end

%%% DBZP BEGIN %%%
N = length(xt);
Nfd = FreqPoints;
Nc = floor(N / Nfd);

% PREALLOCATE FOR SPEED
GPSSpace = zeros(Nfd,Nc*Nfd);

% DIVIDE INPUT SIGNAL INTO Nc-Sized BLOCKS
SigBlocks = zeros(Nfd,Nc);
CaBlocks  = zeros(Nfd,Nc);
for i = 1:Nfd
    Start = 1+ (i-1)*Nc;
    Stop = Start + Nc - 1; 
    SigBlocks(i,:) = xt(Start:Stop);
    CaBlocks(i,:) = C(Start:Stop);
end


 % CREATE DOUBLE-SIZED BLOCKS
 DblSigBlocks = zeros(Nfd,2*Nc);
 DblCaBlocks  = zeros(Nfd,2*Nc);
 
 for i = 1:Nfd

        % GPS INPUT SIGNAL
        if(i == Nfd)
             %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
             % Last Block Circular to First %
             DblSigBlocks(i,:) = [SigBlocks(i,:),SigBlocks(1,:)];
        else
            %%%%%%%%%%%%%%%%%%%%%%
            % Consecutive Blocks %
            DblSigBlocks(i,:) = [SigBlocks(i,:),SigBlocks(i+1,:)];
        end

 end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Add Option to Cancel Computation %
h = waitbar(0,'1','Name','Computation in progress, please wait ...','CreateCancelBtn','setappdata(gcbf,''canceling'',1)');
h.Name = 'GPS DBZP';
setappdata(h,'canceling',0);
 
%%%%%%%%%%%%%%%%%%%%%%%%
% Fix Closing on Error %
try
    % LOOP ACROSS ALL SHIFTS IN A CA DATA BLOCK
    for k = 1:Nfd

        % CANCEL ON REQUEST
        if getappdata(h,'canceling')
            break
        end

        for i = 1:Nfd
            % PRN SIGNAL
            DblCaBlocks(i,:) = [CaBlocks(i,:),zeros(1,Nc)];
        end

        % SHIFT PRN
        CaBlocks = circshift(CaBlocks,[1,0]);

        % PARTIAL CIRCULAR CORRELATION
        for i = 1:Nfd
            PartialCora = ifft(fft(DblSigBlocks(i,:),Nfd).*conj(fft(DblCaBlocks(i,:),Nfd)));
            PartialCora = PartialCora(1:Nc); % Only the first half
            Start = 1 + (k-1)*Nc;
            Stop = Start + Nc-1;
            GPSSpace(i,Start:Stop) = PartialCora;
        end

        % SHOW STATUS (INTEGRATION STATUS)
        waitbar(k / Nfd, h, sprintf('%d//%d', k, Nfd));

    end
catch ME
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Delete WaitBar and Display Error %
    delete(h);
    error(ME.message);
end

%%%%%%%%%%%%%%%%%
% Close Waitbar %
delete(h);

%%%%%%%%%%%%%%%%%%%%
% Column-Wise FFTs %
GPSSpace = fftshift(abs(fft(GPSSpace,Nfd,1)),1);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Smooth Data with Gaussian %
GaussSize = 5;
Sigma = 1;
h = fspecial('gaussian',GaussSize,Sigma);
GPSSpace = imfilter(GPSSpace,h);

end

