%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% @brief: Calculates the Frequency-Delay Space
% xt ...         Input GPS Signal
% C ...          PRN Sequence to decode with
% FStart ...     Doppler Start Freq.
% FStop ...      Doppler Stop Freq.
% Sweep ...      Doppler Sweeping Frequency
% t ... Time axis used for generating the GPS signal
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [ GPSSpace ] = Compute_GNSS_CPU_FFT( xt,C,FStart,FStop,Sweep,t )


%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Check for Invalid Input %
if(length(xt) ~= length(C))
    error('Please, check input signal length and try again');
end

N = length(xt);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Add Option to Cancel Computation %
h = waitbar(0,'1','Name','Computation in progress, please wait ...','CreateCancelBtn','setappdata(gcbf,''canceling'',1)');
h.Name = 'GPS FFT Based Correllation';
setappdata(h,'canceling',0);

FreqPoints  = (abs(FStart) + abs(FStop)) / Sweep + 1;  % Calculate the number of Freq. points
GPSSpace = zeros(FreqPoints,N);                        % Preallocate output for speed
Doppler = FStart:Sweep:FStop;                          % Generate Doppler Frequencies

%%%%%%%%%%%%%%%%%%%%%%%%
% Fix Closing on Error %
try
    for i = 1:FreqPoints

        if getappdata(h,'canceling')
            break
        end

        xtfd = xt.*exp(-1i*2*pi*Doppler(i).*t);      % Simulate a specific doppler
        Cora = abs(ifft(fft(xtfd).*conj(fft(C))));   % Correllation via FFT
        GPSSpace(i,:) = Cora;                        % Save data

        %%%%%%%%%%%%%%%%%
        % Show Progress %
        waitbar(i / FreqPoints,h,sprintf('%d//%d',i,FreqPoints));

    end
catch ME
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Delete WaitBar and Display Error %
    delete(h);
    error(ME.message);
end

%%%%%%%%%%%%%%%%%
% Close WaitBar %
delete(h);

%%%%%%%%%%%%%%%%%%
% Convert To ABS %
GPSSpace = abs(GPSSpace);

end

