%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% @brief: Calculates the Frequency-Delay Space
% xt ...        Input GPS Signal
% C ...         PRN Sequence to decode with
% FStart ...    Doppler Start Freq.
% FStop ...     Doppler Stop Freq.
% Sweep ...     Doppler Sweeping Frequency
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [ GPSSpace ] = Compute_GNSS_CPU_ML( xt,C,FStart,FStop,Sweep,Fsa )


%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Check for Invalid Input %
if(length(xt) ~= length(C))
    error('Please, check input signal length and try again');
end

N = length(xt);
FDInterv = abs(FStart) + abs(FStop);
F = FDInterv / Sweep + 1;


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Add Option to Cancel Computation %
h = waitbar(0,'1','Name','Computation in progress, please wait ...','CreateCancelBtn','setappdata(gcbf,''canceling'',1)');
h.Name = 'GPS ML';
setappdata(h,'canceling',0);

%%%%%%%%%%%%%%%%%%%%%%%%
% Fix Closing on Error %
try

    % INIT VARIABLES
    PRN = zeros(F,N);
    GPSSpace = zeros(F,N);

    % CREATE X & Y INDEXES
    Tix = 0:1:N-1;
    Tiy = 0:1:F-1;
    Tiy = Tiy';
    Tiy = repmat(Tiy,[1,N]);

    % FIX PRN INDEXING
    Cx = [C,C];

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Precalculate Constants to speedup computation % 
    ConstX = -2*pi*(1/Fsa)*(Sweep*Tiy + FStart);

    for k = 1:N % Integration
        if getappdata(h,'canceling')
            break
        end

        PrnInd = k - Tix + N;
        PRNx = Cx(PrnInd);
        for i = 1:F
            PRN(i,:) = PRNx;
        end
        % Repmat is slower than code above by 0.01*N seconds
        % PRN = repmat(PRNx,[F,1]);
        GPS = ones(F,N).*xt(k);
        GPSSpace = GPSSpace + PRN.*GPS.*exp(1i*ConstX*(k));

        %%%%%%%%%%%%%%%%%
        % Show Progress %
        waitbar(k / N,h,sprintf('%d//%d',k,N));
    end

catch ME
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Delete WaitBar and Display Error %
    delete(h);
    error(ME.message);
end

%%%%%%%%%%%%%%%%%
% Close WaitBar %
delete(h);

%%%%%%%%%%%%%%%%%%
% Convert To ABS %
GPSSpace = abs(GPSSpace);
    
end

