clear all;
close all;
clc;

%%%%%%%%%%%%%%%%%%%%%%%%%%
% Get Display Resolution %
Displ = groot();
ScW = Displ.ScreenSize(3);
ScH = Displ.ScreenSize(4);
FiW = 1024;
FiH = 720;
Cls = perms([0,102,204])./255;
Cl_Orange = 1;
Cl_Orchid = 2;
Cl_Green  = 3;
Cl_Violet = 4;
Cl_Tyrkys = 5;
Cl_Blue   = 6;
Lw = 2; 


N = 7;                        % Number of visible satellites [-]
S = zeros(N+1,3);             % Matrix with coordinatess of all satellites
Ranges = zeros(1,N+1);        % Measured pseudodistances [m] 
EarthRad = 6378000;           % Average Earth radius[m]
SatAlt =  20350000;           % Satellite orbitting altitude [m]
RecLatitude = 151;            % Receiver's Latitude [deg]
RecLongtitude = 65;           % Receiver's Logtitude [deg]
RecAltitude = 6190;           % Receiver's Altitude [m]
SatAng = 360/N;               % Angle between Satellites [degrees]
Angles = 0:SatAng:360;        % Angle of each satellite in x-y plane [degrees]
SatDist = EarthRad + SatAlt;  % Satellite Dist from Earh Center to Orbit around the equator [m]
S(N+1,:) =  [0,0,SatDist];    % One satellite direcly above the North Pole
Pos = [0,0,0];                % Initial coordinates for position calculation (Initial Guess) [x,y,z]m
MaxIter = 17;                 % Maximum number of iterations allowed [-]
Threshold = 0.1;              % Threshold to stop iterating [m]
N = N+1;                      % One above the receiver - Do not change
SCor = 1700000;               % Correction for Text Displaying in Constellation chart

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Transform To Spherical Coordinates %
[ReceiverPos(1),ReceiverPos(2),ReceiverPos(3)] = sph2cart(deg2rad(RecLongtitude),deg2rad(RecLatitude),EarthRad+RecAltitude);

if(N < 4)
    error('Need at least 4 Satellites!');
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Add Satellite Position and Ranges %
for i = 1:N-1
    x = cos(deg2rad(Angles(i))) * SatDist;
    y = sin(deg2rad(Angles(i))) * SatDist;
    z = 0;
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%
    % Save Satellites Coords %
    S(i,:) = [x,y,z]; 
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Calculate Distance to Satellites %
    Ranges(i) = sqrt( (x - ReceiverPos(1))^2 + (y-ReceiverPos(2))^2 + (z - ReceiverPos(3))^2 );
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Range for the Last Satellite %
Ranges(N) = sqrt( (S(end,1) - ReceiverPos(1))^2 + (S(end,2) - ReceiverPos(2))^2 + (S(end,3) - ReceiverPos(3))^2 );

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Display Satellite Constellation %
figure('name','Constellation','position',[ScW/2 - FiW/2,ScH/2 - FiH/2,FiW,FiH]);
for i = 1:N
    x = S(i,1);
    y = S(i,2);
    plot(x,y,'linestyle','none','Marker','o','MarkerSize',13,'MarkerFaceColor',[0.8,0.8,0.8],'MarkerEdgeColor',Cls(Cl_Green,:));
    hold on;grid on;
    text(x+SCor,y-SCor,horzcat('S',num2str(i)));
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Add Receiver position to the plot %
plot(ReceiverPos(1),ReceiverPos(2),'linestyle','none','Marker','o','MarkerSize',5,'MarkerFaceColor','black','MarkerEdgeColor','black');
text(ReceiverPos(1)+SCor,ReceiverPos(2)-SCor,'R');

%%%%%%%%%%%%%%%%%%%%%%%%%
% Add Earth to the Plot %
x = linspace(-EarthRad,EarthRad,100);
y = sqrt( EarthRad^2 - x.^2);
plot(x,+y,'linestyle','--','color','black','linewidth',Lw);
plot(x,-y,'linestyle','--','color','black','linewidth',Lw);
title('EARTH''S Z-PROJECTION');
axis equal;

A = zeros(N,4);      % Matrix of direction cosines
Counter = 1;         % Iteration counter
PosOrig = Pos;       % Save Initial Position
while(1)

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % For All Visible Satellites %
    % Pos ... Current Estimated Position
    RangeDiff = zeros(N,1);
    for i = 1:N
        
        % MATRIX OF DIRECTION COSINES
        SatelliteDist = sqrt(sum( (Pos-S(i,:)).^2 )); % Distance To Given Satellite
        A(i,:) = [(Pos - S(i,:))/SatelliteDist,1];    % Normalized Disatnces
        
        % CHANGE IN RANGE
        RangeDiff(i) = (Ranges(i) - SatelliteDist);

    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Solve System of Linear Equations
    % Note: This is Generally a Linear System Solver
    % Could be also computed with Jacobi / Gauss-Seidel or SOR Method
    % Least Squares Solution for Equally Weighted data:
    % x' = (At*A)^(-1) * At *b
    % https://gps.alaska.edu/jeff/Classes/GEOS655/Lecture02_GPS_part1_pseudorange.pdf
    Temp = inv(transpose(A)*A);
    PosEst = Temp*transpose(A)*RangeDiff;
    
    PosEst = PosEst(1:3)';
    PosCor(Counter,:) = PosEst;
    
    % POSITION AND DELUSION OF PRECISION (DOP)
    Pos = Pos + PosEst;
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Note: Could be Computed at the End %
    PDOP = sqrt(trace(Temp) - Temp(end,end));
    HDOP = sqrt(trace(Temp) - Temp(end,end) - Temp(end-1,end-1));
    VDOP = sqrt(Temp(3,3));
    
    %%%%%%%%%%%%%%%%%%%%%%%%%
    % Convergence Criterion %
    DistThreshold = sqrt(sum((PosEst).^2));
    if(DistThreshold < Threshold)
        disp(horzcat('Stopped after: ',num2str(Counter),' Iterations.'));
        break;
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Check for Maximum Amount of Iterations %
    Counter = Counter + 1;
    if(Counter == MaxIter)
        warning('Maximum number of Iterations reached!');
        break;
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Calculate Partial Corrections During Processing %
Nc = length(PosCor(:,1));
for i = 1:Nc
    Corrections(i) = sqrt(sum( PosCor(i,:).^2 ));
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Calculate Final Distance Error %
DistError = sqrt(sum(   (ReceiverPos - Pos).^2 ));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Display Corrections during processing %
%figure('name','Position Corrections','position',[ScW/2 - FiW/2,ScH/2 - FiH/2,FiW,FiH]);
%plot(1:Nc,Corrections,'color',Cls(Cl_Orchid,:),'linewidth',Lw);
%grid on;xlabel('Iteration [-]');
%ylabel('Correction [m]');
%title(['Iteration Corrections with Init. Pos at X: ',num2str(PosOrig(1)),' Y: ',num2str(PosOrig(2)),' Z: ',num2str(PosOrig(3))]);

disp(['Calculated Position | ','X: ',num2str(Pos(1)),' Y: ',num2str(Pos(2)),' Z: ',num2str(Pos(3))]);
disp(['Correct    Position | ','X: ',num2str(ReceiverPos(1)),' Y: ',num2str(ReceiverPos(2)),' Z: ',num2str(ReceiverPos(3))]);
disp(['Asolute Distance Error: ',num2str(DistError),' m']);
disp(['Position   DOP: ',num2str(PDOP)]);
disp(['Horizontal DOP: ',num2str(HDOP)]);
disp(['Vertical   DOP: ',num2str(VDOP)]);



